package dev.fitko.fitconnect.core.crypto;

import com.nimbusds.jose.Algorithm;
import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.KeyOperation;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.util.Base64;
import dev.fitko.fitconnect.api.exceptions.internal.KeyGenerationException;
import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.cert.X509v3CertificateBuilder;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.jcajce.JcaX509v3CertificateBuilder;
import org.bouncycastle.operator.ContentSigner;
import org.bouncycastle.operator.OperatorCreationException;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;

import java.math.BigInteger;
import java.security.InvalidParameterException;
import java.security.KeyPair;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPublicKey;
import java.time.Duration;
import java.time.Instant;
import java.util.Date;
import java.util.List;
import java.util.Set;

public class KeyGenerator {

    /**
     * Generate a set of public and private key
     *
     * @param keySize modulus length in bits
     * @return pair of public and private key
     */
    public KeyPair buildKeyPair(final int keySize) {
        try {
            final java.security.KeyPairGenerator keyPairGenerator = java.security.KeyPairGenerator.getInstance("RSA");
            keyPairGenerator.initialize(keySize);
            return keyPairGenerator.generateKeyPair();
        } catch (final NoSuchAlgorithmException | InvalidParameterException e) {
            throw new KeyGenerationException(e.getMessage(), e);
        }
    }

    /**
     * Build a JWK with the given parameters.
     *
     * @param keyPair            pair of public and private key
     * @param keyId              unique identifier
     * @param keyOperation       key operation that determines the usage of the key
     * @param algorithm          encryption algorithm
     * @param includeCertificate add certificate x5c chain
     * @return JWK
     */
    public JWK buildJWK(final KeyPair keyPair, final String keyId, final KeyOperation keyOperation, final Algorithm algorithm, final boolean includeCertificate) {
        final RSAKey.Builder keyBuilder = new RSAKey.Builder((RSAPublicKey) keyPair.getPublic())
                .privateKey(keyPair.getPrivate())
                .keyID(keyId)
                .keyOperations(Set.of(keyOperation))
                .algorithm(algorithm);

        if (includeCertificate) {
            keyBuilder.x509CertChain(getX509CertChain(keyPair));
        }

        return keyBuilder.build();
    }

    private List<Base64> getX509CertChain(final KeyPair keyPair) {
        final X509Certificate cert = getX509Certificate(keyPair);
        try {
            return RSAKey.parse(cert).getX509CertChain();
        } catch (final JOSEException e) {
            throw new KeyGenerationException(e.getMessage(), e);
        }
    }

    private X509Certificate getX509Certificate(final KeyPair keyPair) {
        final Instant now = Instant.now();
        final Date notBefore = Date.from(now);
        final Date notAfter = Date.from(now.plus(Duration.ofDays(365 * 10)));

        try {
            final ContentSigner contentSigner = new JcaContentSignerBuilder("SHA512withRSA").build(keyPair.getPrivate());
            final X500Name x500Name = new X500Name("CN=FIT-Connect");
            final X500Name x500Subject = new X500Name("C=TestCert");
            final BigInteger serialNr = BigInteger.valueOf(now.toEpochMilli());
            final X509v3CertificateBuilder certificateBuilder = new JcaX509v3CertificateBuilder(x500Name, serialNr, notBefore, notAfter, x500Subject, keyPair.getPublic());
            return new JcaX509CertificateConverter().getCertificate(certificateBuilder.build(contentSigner));
        } catch (final CertificateException | OperatorCreationException e) {
            throw new KeyGenerationException(e.getMessage(), e);
        }
    }
}
