package dev.fitko.fitconnect.core.schema;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.domain.schema.SchemaResources;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectInitialisationException;
import dev.fitko.fitconnect.api.exceptions.internal.SchemaNotFoundException;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import dev.fitko.fitconnect.core.http.HttpClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.InvalidPathException;
import java.nio.file.Path;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class SchemaResourceProvider implements SchemaProvider {

    private static final Logger LOGGER = LoggerFactory.getLogger(SchemaResourceProvider.class);

    private static final Pattern ALLOWED_SCHEMA_PATTERN = Pattern.compile("1.\\d+.\\d+");
    private static final ObjectMapper MAPPER = new ObjectMapper();

    private final Map<URI, String> setSchemas = new HashMap<>();

    private final Map<URI, String> metadataSchemas = new HashMap<>();
    private final Map<URI, String> destinationSchemas = new HashMap<>();
    private final Map<URI, String> submissionDataSchemas = new HashMap<>();

    private final HttpClient httpClient;


    public SchemaResourceProvider(final HttpClient httpClient, final SchemaResources schemaResources) {

        this.httpClient = httpClient;

        populateSetSchemas(schemaResources.getSetSchemaPaths());
        populateMetadataSchemas(schemaResources.getMetadataSchemaPaths());
        populateDestinationSchemas(schemaResources.getDestinationSchemaPaths());

        if (schemaResources.getSubmissionDataSchemaPath() == null) {
            populateSubmissionDataSchemas(schemaResources.getSubmissionDataSchemas());
        }else{
            populateSubmissionDataSchemas(schemaResources.getSubmissionDataSchemaPath());
        }

        LOGGER.info("Initialised SDK schemas");
    }

    private void populateMetadataSchemas(final List<String> metadataSchemaPaths) {
        getResourceFiles(metadataSchemaPaths).forEach(this::addMetadataSchema);
    }

    private void populateSetSchemas(final List<String> setSchemaPaths) {
        getResourceFiles(setSchemaPaths).forEach(this::addSetSchema);
    }

    private void populateDestinationSchemas(final List<String> destinationSchemaPaths) {
        getResourceFiles(destinationSchemaPaths).forEach(this::addDestinationSchema);
    }

    private void populateSubmissionDataSchemas(final Map<String, String> dataSchemas) {
        dataSchemas.forEach((schemaKey, schemaPath) -> submissionDataSchemas.putIfAbsent(getUriFromKey(schemaKey), readPathToString(Path.of(schemaPath))));
    }

    private void populateSubmissionDataSchemas(String submissionDataSchemaPath) {
        if (submissionDataSchemaPath != null && !submissionDataSchemaPath.isEmpty()) {
            getDataSchemaPathsFromDir(submissionDataSchemaPath).stream()
                    .map(SchemaResourceProvider::readPathToString)
                    .forEach(this::addSubmissionDataSchema);
        }
    }

    private static String readPathToString(final String schemaPath) {
        try {
            return Files.readString(Path.of(schemaPath));
        } catch (final IOException | NullPointerException e) {
            throw new FitConnectInitialisationException("Reading schema " + schemaPath + " failed", e);
        }
    }

    @Override
    public boolean isAllowedSetSchema(final URI schemaUri) {
        return schemaVersionMatchesPattern(schemaUri, ALLOWED_SCHEMA_PATTERN);
    }

    @Override
    public boolean isAllowedMetadataSchema(final URI schemaUri) {
        return metadataSchemas.containsKey(schemaUri);
    }

    @Override
    public String loadSetSchema(URI schemaUri) throws SchemaNotFoundException {
        final String schema = setSchemas.get(schemaUri);
        if (schema == null) {
            throw new SchemaNotFoundException("SET schema " + schemaUri.toString() + " is not available.");
        }
        return schema;
    }

    @Override
    public String loadMetadataSchema(final URI schemaUri) throws SchemaNotFoundException {
        final String schema = metadataSchemas.get(schemaUri);
        if (schema == null) {
            throw new SchemaNotFoundException("Metadata schema " + schemaUri.toString() + " is not available.");
        }
        return schema;
    }

    @Override
    public String loadDestinationSchema(final URI schemaUri) throws SchemaNotFoundException {
        final String schema = destinationSchemas.get(schemaUri);
        if (schema == null) {
            throw new SchemaNotFoundException("Destination schema " + schemaUri.toString() + " is not available.");
        }
        return schema;
    }

    @Override
    public String loadSubmissionDataSchema(final URI schemaUri) throws SchemaNotFoundException {
        final String schema = submissionDataSchemas.get(schemaUri);
        if (schema != null) {
            return schema;
        }
        LOGGER.debug("Schema " + schemaUri + " is not available as local file, loading remote version.");

        if (!schemaUri.getScheme().equals("https")) {
            throw new SchemaNotFoundException("Fetching schema " + schemaUri + " from remote was skipped, since the URI does not support HTTPS.");
        }

        try {
            return httpClient.get(schemaUri.toString(), String.class).getBody();
        } catch (final Exception exception) {
            throw new SchemaNotFoundException("Submission data schema " + schemaUri + " is not available.", exception);
        }
    }

    private void addSetSchema(final String schema) {

        readIdFromSchema(schema).ifPresentOrElse(id -> setSchemas.put(id, schema), () -> LOGGER.warn("File '" + schema + "' does not provide a valid set schema and will be ignored."));
    }

    private void addMetadataSchema(final String schema) {

        readIdFromSchema(schema).ifPresentOrElse(id -> metadataSchemas.put(id, schema), () -> LOGGER.warn("File '" + schema + "' does not provide a valid metadata schema and will be ignored."));
    }

    private void addDestinationSchema(final String schema) {

        readIdFromSchema(schema).ifPresentOrElse(id -> destinationSchemas.put(id, schema), () -> LOGGER.warn("File '" + schema + "' does not provide a valid destination schema and will be ignored."));
    }

    private void addSubmissionDataSchema(final String schema) {

        readIdFromSchema(schema).ifPresentOrElse(
                id -> submissionDataSchemas.put(id, schema),
                () -> LOGGER.warn("File '" + schema + "' does not provide a valid submission data schema and will be ignored."));
    }

    private Optional<URI> readIdFromSchema(final String schema) {
        try {
            return Optional.of(URI.create(MAPPER.readTree(schema).get("$id").asText()));
        } catch (final JsonProcessingException e) {
            return Optional.empty();
        }
    }

    private List<String> getResourceFiles(final List<String> schemaFiles) {
        return schemaFiles.stream().map(SchemaResourceProvider::readResourceToString).collect(Collectors.toList());
    }

    private static String readPathToString(final Path schemaPath) {
        try {
            return Files.readString(schemaPath);
        } catch (final IOException | NullPointerException e) {
            throw new FitConnectInitialisationException("Reading schema " + schemaPath.toString() + " failed", e);
        }
    }

    private static String readResourceToString(final String schemaFile) {
        try (final InputStream is = SchemaResourceProvider.class.getResourceAsStream(schemaFile); final BufferedReader reader = new BufferedReader(new InputStreamReader(is))) {
            return reader.lines().collect(Collectors.joining(System.lineSeparator()));
        } catch (final Exception e) {
            throw new FitConnectInitialisationException("Could not read schema file " + schemaFile, e);
        }
    }

    private static List<String> getDataSchemaPathsFromDir(final String submissionDataSchemaPath) {
        if (submissionDataSchemaPath == null) {
            return Collections.emptyList();
        }
        try (final Stream<Path> paths = Files.walk(Path.of(submissionDataSchemaPath))) {
            return paths.filter(Files::isRegularFile).map(Path::toString).collect(Collectors.toList());
        } catch (final IOException | InvalidPathException e) {
            throw new FitConnectInitialisationException("Could not read submission data schemas from path " + submissionDataSchemaPath, e);
        }
    }

    private boolean schemaVersionMatchesPattern(final URI schemaUri, final Pattern pattern) {
        final String schemaVersion = schemaUri.getPath().split("/")[3];
        return pattern.matcher(schemaVersion).matches();
    }

    private static URI getUriFromKey(final String schemaKey) {
        try {
            return URI.create(schemaKey);
        } catch (final Exception e) {
            LOGGER.error("Could not map data schema {} to a valid URI", schemaKey);
            throw new FitConnectInitialisationException(e.getMessage(), e);
        }
    }
}
